﻿using System.Collections.Generic;
using EonaCat.Quic.Constants;
using EonaCat.Quic.Events;
using EonaCat.Quic.Exceptions;
using EonaCat.Quic.Helpers;
using EonaCat.Quic.Infrastructure;
using EonaCat.Quic.Infrastructure.Frames;
using EonaCat.Quic.Infrastructure.PacketProcessing;
using EonaCat.Quic.Infrastructure.Packets;
using EonaCat.Quic.Infrastructure.Settings;
using EonaCat.Quic.InternalInfrastructure;
using EonaCat.Quic.Streams;

namespace EonaCat.Quic.Connections;
// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

public class QuicConnection
{
    private readonly NumberSpace _numberSpace = new();
    private readonly PacketWireTransfer _pwt;
    private readonly Dictionary<ulong, QuicStream> _streams;

    private ulong _currentTransferRate;
    private string _lastError;
    private ConnectionState _state;

    internal QuicConnection(ConnectionData connection)
    {
        _currentTransferRate = 0;
        _state = ConnectionState.Open;
        _lastError = string.Empty;
        _streams = new Dictionary<ulong, QuicStream>();
        _pwt = connection.PWT;

        ConnectionId = connection.ConnectionId;
        PeerConnectionId = connection.PeerConnectionId;
        // Also creates a new number space
        PacketCreator = new PacketCreator(ConnectionId, PeerConnectionId);
        MaxData = QuicSettings.MaxData;
        MaxStreams = QuicSettings.MaximumStreamId;
    }

    public IntegerParts ConnectionId { get; }
    public IntegerParts PeerConnectionId { get; }

    public PacketCreator PacketCreator { get; }
    public ulong MaxData { get; private set; }
    public ulong MaxStreams { get; private set; }

    public StreamOpenedEvent OnStreamOpened { get; set; }
    public ConnectionClosedEvent OnConnectionClosed { get; set; }

    /// <summary>
    ///     Creates a new stream for sending/receiving data.
    /// </summary>
    /// <param name="type">Type of the stream (Uni-Bidirectional)</param>
    /// <returns>A new stream instance or Null if the connection is terminated.</returns>
    public QuicStream CreateStream(StreamType type)
    {
        var streamId = _numberSpace.Get();
        if (_state != ConnectionState.Open)
        {
            return null;
        }

        var stream = new QuicStream(this, new StreamId(streamId, type));
        _streams.Add(streamId, stream);

        return stream;
    }

    public QuicStream ProcessFrames(List<Frame> frames)
    {
        QuicStream stream = null;

        foreach (var frame in frames)
        {
            if (frame.Type == 0x01)
            {
                OnRstStreamFrame(frame);
            }

            if (frame.Type == 0x04)
            {
                OnRstStreamFrame(frame);
            }

            if (frame.Type >= 0x08 && frame.Type <= 0x0f)
            {
                stream = OnStreamFrame(frame);
            }

            if (frame.Type == 0x10)
            {
                OnMaxDataFrame(frame);
            }

            if (frame.Type == 0x11)
            {
                OnMaxStreamDataFrame(frame);
            }

            if (frame.Type >= 0x12 && frame.Type <= 0x13)
            {
                OnMaxStreamFrame(frame);
            }

            if (frame.Type == 0x14)
            {
                OnDataBlockedFrame(frame);
            }

            if (frame.Type >= 0x1c && frame.Type <= 0x1d)
            {
                OnConnectionCloseFrame(frame);
            }
        }

        return stream;
    }

    public void IncrementRate(int length)
    {
        _currentTransferRate += (uint)length;
    }

    public bool MaximumReached()
    {
        if (_currentTransferRate >= MaxData)
        {
            return true;
        }

        return false;
    }

    private void OnConnectionCloseFrame(Frame frame)
    {
        var ccf = (ConnectionCloseFrame)frame;
        _state = ConnectionState.Draining;
        _lastError = ccf.ReasonPhrase;

        OnConnectionClosed?.Invoke(this);
    }

    private void OnRstStreamFrame(Frame frame)
    {
        var rsf = (ResetStreamFrame)frame;
        if (_streams.ContainsKey(rsf.StreamId))
        {
            // Find and reset the stream
            var stream = _streams[rsf.StreamId];
            stream.ResetStream(rsf);

            // Remove the stream from the connection
            _streams.Remove(rsf.StreamId);
        }
    }

    private QuicStream OnStreamFrame(Frame frame)
    {
        QuicStream stream;

        var sf = (StreamFrame)frame;
        StreamId streamId = sf.StreamId;

        if (_streams.ContainsKey(streamId.Id) == false)
        {
            stream = new QuicStream(this, streamId);

            if ((ulong)_streams.Count < MaxStreams)
            {
                _streams.Add(streamId.Id, stream);
            }
            else
            {
                SendMaximumStreamReachedError();
            }

            OnStreamOpened?.Invoke(stream);
        }
        else
        {
            stream = _streams[streamId.Id];
        }

        stream.ProcessData(sf);

        return stream;
    }

    private void OnMaxDataFrame(Frame frame)
    {
        var sf = (MaxDataFrame)frame;
        if (sf.MaximumData.Value > MaxData)
        {
            MaxData = sf.MaximumData.Value;
        }
    }

    private void OnMaxStreamDataFrame(Frame frame)
    {
        var msdf = (MaxStreamDataFrame)frame;
        StreamId streamId = msdf.StreamId;
        if (_streams.ContainsKey(streamId.Id))
        {
            // Find and set the new maximum stream data on the stream
            var stream = _streams[streamId.Id];
            stream.SetMaximumStreamData(msdf.MaximumStreamData.Value);
        }
    }

    private void OnMaxStreamFrame(Frame frame)
    {
        var msf = (MaxStreamsFrame)frame;
        if (msf.MaximumStreams > MaxStreams)
        {
            MaxStreams = msf.MaximumStreams.Value;
        }
    }

    private void OnDataBlockedFrame(Frame frame)
    {
        TerminateConnection();
    }

    public QuicStream OpenStream()
    {
        QuicStream stream = null;

        while (stream == null)
        {
            var packet = _pwt.ReadPacket();
            if (packet is ShortHeaderPacket shp)
            {
                stream = ProcessFrames(shp.GetFrames());
            }
        }

        return stream;
    }

    /// <summary>
    ///     Client only!
    /// </summary>
    /// <returns></returns>
    internal void ReceivePacket()
    {
        var packet = _pwt.ReadPacket();

        if (packet is ShortHeaderPacket shp)
        {
            ProcessFrames(shp.GetFrames());
        }

        // If the connection has been closed
        if (_state == ConnectionState.Draining)
        {
            if (string.IsNullOrWhiteSpace(_lastError))
            {
                _lastError = "Protocol error";
            }

            TerminateConnection();

            throw new ConnectionException(_lastError);
        }
    }

    internal bool SendData(Packet packet)
    {
        return _pwt.SendPacket(packet);
    }

    internal void TerminateConnection()
    {
        _state = ConnectionState.Draining;
        _streams.Clear();

        ConnectionPool.RemoveConnection(ConnectionId);
    }

    internal void SendMaximumStreamReachedError()
    {
        var packet =
            PacketCreator.CreateConnectionClosePacket(ErrorCode.STREAM_LIMIT_ERROR, 0x00,
                ErrorConstants.MaxNumberOfStreams);
        Send(packet);
    }

    /// <summary>
    ///     Used to send protocol packets to the peer.
    /// </summary>
    /// <param name="packet"></param>
    /// <returns></returns>
    internal bool Send(Packet packet)
    {
        // Encode the packet
        var data = packet.Encode();

        // Increment the connection transfer rate
        IncrementRate(data.Length);

        // If the maximum transfer rate is reached, send FLOW_CONTROL_ERROR
        if (MaximumReached())
        {
            packet = PacketCreator.CreateConnectionClosePacket(ErrorCode.FLOW_CONTROL_ERROR, 0x00,
                ErrorConstants.MaxDataTransfer);

            TerminateConnection();
        }

        // Ignore empty packets
        if (data == null || data.Length <= 0)
        {
            return true;
        }

        var result = _pwt.SendPacket(packet);

        return result;
    }
}