﻿using System;
using System.Collections.Generic;
using System.Linq;
using EonaCat.Quic.Connections;
using EonaCat.Quic.Constants;
using EonaCat.Quic.Events;
using EonaCat.Quic.Exceptions;
using EonaCat.Quic.Helpers;
using EonaCat.Quic.Infrastructure;
using EonaCat.Quic.Infrastructure.Frames;
using EonaCat.Quic.Infrastructure.Settings;

namespace EonaCat.Quic.Streams;
// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

/// <summary>
///     Virtual multiplexing channel.
/// </summary>
public class QuicStream
{
    private readonly QuicConnection _connection;
    private readonly SortedList<ulong, byte[]> _data = new();
    private ulong _currentTransferRate;
    private ulong _maximumStreamData;
    private ulong _sendOffset;

    public QuicStream(QuicConnection connection, StreamId streamId)
    {
        StreamId = streamId;
        Type = streamId.Type;

        _maximumStreamData = QuicSettings.MaxStreamData;
        _currentTransferRate = 0;
        _sendOffset = 0;

        _connection = connection;
    }

    public StreamState State { get; set; }
    public StreamType Type { get; set; }
    public StreamId StreamId { get; }

    public StreamDataReceivedEvent OnStreamDataReceived { get; set; }

    public byte[] Data => _data.SelectMany(v => v.Value).ToArray();

    public bool Send(byte[] data)
    {
        if (Type == StreamType.ServerUnidirectional)
        {
            throw new StreamException("Cannot send data on unidirectional stream.");
        }

        _connection.IncrementRate(data.Length);

        var numberOfPackets = data.Length / QuicSettings.PMTU + 1;
        var leftoverCarry = data.Length % QuicSettings.PMTU;

        for (var i = 0; i < numberOfPackets; i++)
        {
            var eos = false;
            var dataSize = QuicSettings.PMTU;
            if (i == numberOfPackets - 1)
            {
                eos = true;
                dataSize = leftoverCarry;
            }

            var buffer = new byte[dataSize];
            Buffer.BlockCopy(data, (int)_sendOffset, buffer, 0, dataSize);

            var packet = _connection.PacketCreator.CreateDataPacket(StreamId.IntegerValue, buffer, _sendOffset, eos);
            if (i == 0 && data.Length >= QuicSettings.MaxStreamData)
            {
                packet.AttachFrame(new MaxStreamDataFrame(StreamId.IntegerValue, (ulong)(data.Length + 1)));
            }

            if (_connection.MaximumReached())
            {
                packet.AttachFrame(new StreamDataBlockedFrame(StreamId.IntegerValue, (ulong)data.Length));
            }

            _sendOffset += (ulong)buffer.Length;

            _connection.SendData(packet);
        }

        return true;
    }

    /// <summary>
    ///     Client only!
    /// </summary>
    /// <returns></returns>
    public byte[] Receive()
    {
        if (Type == StreamType.ClientUnidirectional)
        {
            throw new StreamException("Cannot receive data on unidirectional stream.");
        }

        while (!IsStreamFull() || State == StreamState.Receive) _connection.ReceivePacket();

        return Data;
    }

    public void ResetStream(ResetStreamFrame frame)
    {
        // Reset the state
        State = StreamState.ResetReceived;
        // Clear data
        _data.Clear();
    }

    public void SetMaximumStreamData(ulong maximumData)
    {
        _maximumStreamData = maximumData;
    }

    public bool CanSendData()
    {
        if (Type == StreamType.ServerUnidirectional || Type == StreamType.ClientUnidirectional)
        {
            return false;
        }

        if (State == StreamState.Receive || State == StreamState.SizeKnown)
        {
            return true;
        }

        return false;
    }

    public bool IsOpen()
    {
        if (State == StreamState.DataReceived || State == StreamState.ResetReceived)
        {
            return false;
        }

        return true;
    }

    public void ProcessData(StreamFrame frame)
    {
        // Do not accept data if the stream is reset.
        if (State == StreamState.ResetReceived)
        {
            return;
        }

        var data = frame.StreamData;
        if (frame.Offset != null)
        {
            _data.Add(frame.Offset.Value, frame.StreamData);
        }
        else
        {
            _data.Add(0, frame.StreamData);
        }

        // Either this frame marks the end of the stream,
        // or fin frame came before the data frames
        if (frame.EndOfStream)
        {
            State = StreamState.SizeKnown;
        }

        _currentTransferRate += (ulong)data.Length;

        // Terminate connection if maximum stream data is reached
        if (_currentTransferRate >= _maximumStreamData)
        {
            var errorPacket = _connection.PacketCreator.CreateConnectionClosePacket(ErrorCode.FLOW_CONTROL_ERROR,
                frame.ActualType, ErrorConstants.MaxDataTransfer);
            _connection.SendData(errorPacket);
            _connection.TerminateConnection();

            return;
        }

        if (State == StreamState.SizeKnown && IsStreamFull())
        {
            State = StreamState.DataReceived;

            OnStreamDataReceived?.Invoke(this, Data);
        }
    }

    public void ProcessStreamDataBlocked(StreamDataBlockedFrame frame)
    {
        State = StreamState.DataReceived;
    }

    private bool IsStreamFull()
    {
        ulong length = 0;

        foreach (var kvp in _data)
        {
            if (kvp.Key > 0 && kvp.Key != length)
            {
                return false;
            }

            length += (ulong)kvp.Value.Length;
        }

        return true;
    }
}