﻿using System;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class SocketTcpServer
    {
        private const int BUFFER_SIZE = 4096;

        private readonly TcpListener _listener;
        private readonly X509Certificate2 _certificate;
        private readonly SslOptions _sslOptions;
        private Task _acceptTask;
        private CancellationTokenSource _cancellationTokenSource;

        public SocketTcpServer(IPAddress ipAddress, int port, X509Certificate2 certificate = null, SslOptions sslOptions = null)
        {
            _listener = new TcpListener(ipAddress, port);
            _certificate = certificate;
            _sslOptions = sslOptions;
        }

        public SocketTcpServer(string ipAddress, int port, X509Certificate2 certificate = null, SslOptions sslOptions = null)
        {
            var address = IPAddress.Parse(ipAddress);
            _listener = new TcpListener(address, port);
            _certificate = certificate;
            _sslOptions = sslOptions;
        }

        public event Action<RemoteInfo> OnConnect;
        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnSend;
        public event Action<RemoteInfo> OnDisconnect;
        public event Action<Exception, string> OnError;

        public Task StartAsync(CancellationToken cancellationToken = default)
        {
            _listener.Start();
            _cancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            _acceptTask = AcceptConnectionsAsync(_cancellationTokenSource.Token);
            return _acceptTask;
        }

        private async Task AcceptConnectionsAsync(CancellationToken cancellationToken)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var tcpClient = await _listener.AcceptTcpClientAsync().ConfigureAwait(false);
                    _ = HandleConnectionAsync(tcpClient, cancellationToken);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                }
            }
        }

        private async Task HandleConnectionAsync(TcpClient tcpClient, CancellationToken cancellationToken)
        {
            var remoteEndpoint = tcpClient.Client.RemoteEndPoint;
            using (tcpClient)
            {
                OnConnect?.Invoke(new RemoteInfo
                {
                    Socket = tcpClient.Client,
                    IsTcp = true,
                    IsIPv6 = remoteEndpoint.AddressFamily == AddressFamily.InterNetworkV6,
                    EndPoint = remoteEndpoint
                });

                Stream stream = tcpClient.GetStream();
                if (_certificate != null)
                {
                    var sslStream = new SslStream(stream, false, ValidateRemoteCertificate, null);
                    try
                    {
                        await sslStream.AuthenticateAsServerAsync(_certificate, _sslOptions.ClientCertificateRequired, _sslOptions.SslProtocol, _sslOptions.CheckCertificateRevocation);
                        stream = sslStream;
                    }
                    catch (AuthenticationException ex)
                    {
                        OnError?.Invoke(ex, $"AuthenticationException: {ex.Message}");
                        return;
                    }
                }

                var buffer = new byte[BUFFER_SIZE];
                while (!cancellationToken.IsCancellationRequested)
                {
                    try
                    {
                        var received = await stream.ReadAsync(buffer, 0, buffer.Length, cancellationToken).ConfigureAwait(false);
                        if (received > 0)
                        {
                            var data = new byte[received];
                            Buffer.BlockCopy(buffer, 0, data, 0, received);
                            OnReceive?.Invoke(new RemoteInfo
                            {
                                Socket = tcpClient.Client,
                                IsTcp = true,
                                IsIPv6 = remoteEndpoint.AddressFamily == AddressFamily.InterNetworkV6,
                                EndPoint = remoteEndpoint,
                                Data = data
                            });
                        }
                        else
                        {
                            break;
                        }
                    }
                    catch (IOException ex)
                    {
                        OnError?.Invoke(ex, $"IOException: {ex.Message}");
                        break;
                    }
                }
            }

            OnDisconnect?.Invoke(new RemoteInfo
            {
                Socket = tcpClient.Client,
                IsTcp = true,
                EndPoint = remoteEndpoint,
                IsIPv6 = remoteEndpoint.AddressFamily == AddressFamily.InterNetworkV6
            });
        }

        public async Task SendToAsync(Socket socket, byte[] data)
        {
            await socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None).ConfigureAwait(false);
            OnSend?.Invoke(new RemoteInfo
            {
                Socket = socket,
                IsTcp = true,
                EndPoint = socket.RemoteEndPoint,
                IsIPv6 = socket.AddressFamily == AddressFamily.InterNetworkV6,
                Data = data
            });
        }

        public async Task StopAsync()
        {
            _cancellationTokenSource.Cancel();
            _listener.Stop();
            if (_acceptTask != null)
            {
                try
                {
                    await _acceptTask.ConfigureAwait(false);
                }
                catch (AggregateException)
                {
                }
            }
        }

        private bool ValidateRemoteCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            if (_sslOptions != null && _sslOptions.CertificateValidationCallback != null)
            {
                return _sslOptions.CertificateValidationCallback(sender, certificate, chain, sslPolicyErrors);
            }
            return sslPolicyErrors == SslPolicyErrors.None;
        }
    }
}
