﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using EonaCat.Logger.Extensions;
using EonaCat.Network;

namespace EonaCat.WebSockets;
// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

public sealed class AsyncWebSocketServer
{
    private const int _none = 0;
    private const int _listening = 1;
    private const int _disposed = 5;
    private readonly AsyncWebSocketServerModuleCatalog _catalog;
    private readonly AsyncWebSocketServerConfiguration _configuration;
    private readonly ConcurrentDictionary<string, AsyncWebSocketSession> _sessions = new();
    private TcpListener _listener;
    private AsyncWebSocketRouteResolver _routeResolver;

    private int _state;

    public AsyncWebSocketServer(int listenedPort, AsyncWebSocketServerModuleCatalog catalog,
        AsyncWebSocketServerConfiguration configuration = null)
        : this(IPAddress.Any, listenedPort, catalog, configuration)
    {
    }

    public AsyncWebSocketServer(IPAddress listenedAddress, int listenedPort, AsyncWebSocketServerModuleCatalog catalog,
        AsyncWebSocketServerConfiguration configuration = null)
        : this(new IPEndPoint(listenedAddress, listenedPort), catalog, configuration)
    {
    }

    public AsyncWebSocketServer(IPEndPoint listenedEndPoint, AsyncWebSocketServerModuleCatalog catalog,
        AsyncWebSocketServerConfiguration configuration = null)
    {
        if (listenedEndPoint == null)
        {
            throw new ArgumentNullException("listenedEndPoint");
        }

        if (catalog == null)
        {
            throw new ArgumentNullException("catalog");
        }

        ListenedEndPoint = listenedEndPoint;
        _catalog = catalog;
        _configuration = configuration ?? new AsyncWebSocketServerConfiguration();

        if (_configuration.BufferManager == null)
        {
            throw new InvalidProgramException("The buffer manager in configuration cannot be null.");
        }

        Initialize();
    }

    public IPEndPoint ListenedEndPoint { get; }
    public bool IsListening => _state == _listening;
    public int SessionCount => _sessions.Count;

    public IEnumerable<string> EnabledExtensions =>
        _configuration.EnabledExtensions != null ? _configuration.EnabledExtensions.Keys : null;

    public IEnumerable<string> EnabledSubProtocols =>
        _configuration.EnabledSubProtocols != null ? _configuration.EnabledSubProtocols.Keys : null;

    private void Initialize()
    {
        _routeResolver = new AsyncWebSocketRouteResolver(_catalog);
    }


    public void Listen()
    {
        var origin = Interlocked.CompareExchange(ref _state, _listening, _none);
        if (origin == _disposed)
        {
            throw new ObjectDisposedException(GetType().FullName);
        }

        if (origin != _none)
        {
            throw new InvalidOperationException("This websocket server has already started.");
        }

        try
        {
            _listener = new TcpListener(ListenedEndPoint);
            ConfigureListener();

            _listener.Start(_configuration.PendingConnectionBacklog);

            Task.Factory.StartNew(async () => { await Accept(); },
                    TaskCreationOptions.LongRunning)
                .Forget();
        }
        catch (Exception ex) when (!ShouldThrow(ex))
        {
        }
    }

    public void Shutdown()
    {
        if (Interlocked.Exchange(ref _state, _disposed) == _disposed)
        {
            return;
        }

        try
        {
            _listener.Stop();
            _listener = null;

            Task.Factory.StartNew(async () =>
                    {
                        try
                        {
                            foreach (var session in _sessions.Values)
                            {
                                await session.Close(WebSocketCloseCode.NormalClosure);
                            }
                        }
                        catch (Exception ex) when (!ShouldThrow(ex))
                        {
                        }
                    },
                    TaskCreationOptions.PreferFairness)
                .Wait();
        }
        catch (Exception ex) when (!ShouldThrow(ex))
        {
        }
    }

    private void ConfigureListener()
    {
        _listener.AllowNatTraversal(_configuration.AllowNatTraversal);
    }

    public bool Pending()
    {
        if (!IsListening)
        {
            throw new InvalidOperationException("The websocket server is not active.");
        }

        // determine if there are pending connection requests.
        return _listener.Pending();
    }

    private async Task Accept()
    {
        try
        {
            while (IsListening)
            {
                var tcpClient = await _listener.AcceptTcpClientAsync();
                Task.Factory.StartNew(async () => { await Process(tcpClient); },
                        TaskCreationOptions.PreferFairness)
                    .Forget();
            }
        }
        catch (Exception ex) when (!ShouldThrow(ex))
        {
        }
        catch (Exception ex)
        {
            NetworkHelper.Logger.Error(ex.FormatExceptionToMessage());
        }
    }

    private async Task Process(TcpClient acceptedTcpClient)
    {
        var session = new AsyncWebSocketSession(acceptedTcpClient, _configuration, _configuration.BufferManager,
            _routeResolver, this);

        if (_sessions.TryAdd(session.SessionKey, session))
        {
            NetworkHelper.Logger.Debug($"New session [{session}].");
            try
            {
                await session.Start();
            }
            catch (Exception ex)
                when (ex is TimeoutException || ex is WebSocketException)
            {
                NetworkHelper.Logger.Error(ex.FormatExceptionToMessage());
            }
            finally
            {
                if (_sessions.TryRemove(session.SessionKey, out var throwAway))
                {
                    NetworkHelper.Logger.Debug($"Close session [{throwAway}].");
                }
            }
        }
    }

    private static bool ShouldThrow(Exception ex)
    {
        return ex is not ObjectDisposedException
               && ex is not InvalidOperationException
               && ex is not SocketException
               && ex is not IOException;
    }

    public async Task SendTextToAsync(string sessionKey, string text)
    {
        if (_sessions.TryGetValue(sessionKey, out var sessionFound))
        {
            await sessionFound.SendTextAsync(text);
        }
        else
        {
            NetworkHelper.Logger.Warn($"Cannot find session [{sessionKey}].");
        }
    }

    public async Task SendTextToAsync(AsyncWebSocketSession session, string text)
    {
        AsyncWebSocketSession sessionFound;
        if (_sessions.TryGetValue(session.SessionKey, out sessionFound))
        {
            await sessionFound.SendTextAsync(text);
        }
        else
        {
            NetworkHelper.Logger.Warn($"Send text data but cannot find session [{session}].");
        }
    }

    public async Task SendBinaryToAsync(string sessionKey, byte[] data)
    {
        await SendBinaryToAsync(sessionKey, data, 0, data.Length);
    }

    public async Task SendBinaryToAsync(string sessionKey, byte[] data, int offset, int count)
    {
        if (_sessions.TryGetValue(sessionKey, out var sessionFound))
        {
            await sessionFound.SendBinaryAsync(data, offset, count);
        }
        else
        {
            NetworkHelper.Logger.Warn($"Cannot find session [{sessionKey}].");
        }
    }

    public async Task SendBinaryToAsync(AsyncWebSocketSession session, byte[] data)
    {
        await SendBinaryToAsync(session, data, 0, data.Length);
    }

    public async Task SendBinaryToAsync(AsyncWebSocketSession session, byte[] data, int offset, int count)
    {
        if (_sessions.TryGetValue(session.SessionKey, out var sessionFound))
        {
            await sessionFound.SendBinaryAsync(data, offset, count);
        }
        else
        {
            NetworkHelper.Logger.Warn($"Send binary data but cannot find session [{session}].");
        }
    }

    public async Task BroadcastTextAsync(string text)
    {
        foreach (var session in _sessions.Values)
        {
            await session.SendTextAsync(text);
        }
    }

    public async Task BroadcastBinaryAsync(byte[] data)
    {
        await BroadcastBinaryAsync(data, 0, data.Length);
    }

    public async Task BroadcastBinaryAsync(byte[] data, int offset, int count)
    {
        foreach (var session in _sessions.Values)
        {
            await session.SendBinaryAsync(data, offset, count);
        }
    }

    public bool HasSession(string sessionKey)
    {
        return _sessions.ContainsKey(sessionKey);
    }

    public AsyncWebSocketSession GetSession(string sessionKey)
    {
        _sessions.TryGetValue(sessionKey, out var session);
        return session;
    }

    public async Task CloseSession(string sessionKey)
    {
        if (_sessions.TryGetValue(sessionKey, out var session))
        {
            await session.Close(WebSocketCloseCode.NormalClosure);
        }
    }
}