﻿using System;
using System.Threading.Tasks;

namespace EonaCat.Network;
// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

/// <summary>
///     EonaCat webserver routes.
/// </summary>
public class EonaCatWebserverRoutes
{
    private readonly EonaCatWebserverSettings _Settings = new();
    private ContentRouteManager _Content = new();
    private ContentRouteHandler _ContentHandler;
    private DynamicRouteManager _Dynamic = new();
    private ParameterRouteManager _Parameter = new();
    private Func<HttpContext, Task> _Preflight;

    private StaticRouteManager _Static = new();

    /// <summary>
    ///     Function to call prior to routing.
    ///     Return 'true' if the connection should be terminated.
    ///     Return 'false' to allow the connection to continue routing.
    /// </summary>
    public Func<HttpContext, Task<bool>> PreRouting = null;

    /// <summary>
    ///     Instantiate the object using default settings.
    /// </summary>
    public EonaCatWebserverRoutes()
    {
        _Preflight = PreflightInternal;
        _ContentHandler = new ContentRouteHandler(_Content);
    }

    /// <summary>
    ///     Instantiate the object using default settings and the specified default route.
    /// </summary>
    public EonaCatWebserverRoutes(EonaCatWebserverSettings settings, Func<HttpContext, Task> defaultRoute)
    {
        settings ??= new EonaCatWebserverSettings();

        if (defaultRoute == null)
        {
            throw new ArgumentNullException(nameof(defaultRoute));
        }

        _Settings = settings;
        _Preflight = PreflightInternal;
        Default = defaultRoute;
        _ContentHandler = new ContentRouteHandler(_Content);
    }

    /// <summary>
    ///     Function to call when a preflight (OPTIONS) request is received.
    ///     Often used to handle CORS.
    ///     Leave null to use the default OPTIONS handler.
    /// </summary>
    public Func<HttpContext, Task> Preflight
    {
        get => _Preflight;
        set
        {
            if (value == null)
            {
                _Preflight = PreflightInternal;
            }
            else
            {
                _Preflight = value;
            }
        }
    }

    /// <summary>
    ///     Content routes; i.e. routes to specific files or folders for GET and HEAD requests.
    /// </summary>
    public ContentRouteManager Content
    {
        get => _Content;
        set
        {
            if (value == null)
            {
                throw new ArgumentNullException(nameof(Content));
            }

            _Content = value;
        }
    }

    /// <summary>
    ///     Handler for content route requests.
    /// </summary>
    public ContentRouteHandler ContentHandler
    {
        get => _ContentHandler;
        set
        {
            if (value == null)
            {
                throw new ArgumentNullException(nameof(ContentHandler));
            }

            _ContentHandler = value;
        }
    }

    /// <summary>
    ///     Static routes; i.e. routes with explicit matching and any HTTP method.
    /// </summary>
    public StaticRouteManager Static
    {
        get => _Static;
        set
        {
            if (value == null)
            {
                throw new ArgumentNullException(nameof(Static));
            }

            _Static = value;
        }
    }

    /// <summary>
    ///     Parameter routes; i.e. routes with parameters embedded in the URL, such as /{version}/api/{id}.
    /// </summary>
    public ParameterRouteManager Parameter
    {
        get => _Parameter;
        set
        {
            if (value == null)
            {
                throw new ArgumentNullException(nameof(Parameter));
            }

            _Parameter = value;
        }
    }

    /// <summary>
    ///     Dynamic routes; i.e. routes with regex matching and any HTTP method.
    /// </summary>
    public DynamicRouteManager Dynamic
    {
        get => _Dynamic;
        set
        {
            if (value == null)
            {
                throw new ArgumentNullException(nameof(Dynamic));
            }

            _Dynamic = value;
        }
    }

    /// <summary>
    ///     Default route; used when no other routes match.
    /// </summary>
    public Func<HttpContext, Task> Default { get; set; }

    private async Task PreflightInternal(HttpContext context)
    {
        context.Response.StatusCode = 200;

        string[] requestedHeaders = null;
        if (context.Request.Headers != null)
        {
            foreach (var curr in context.Request.Headers)
            {
                if (string.IsNullOrEmpty(curr.Key))
                {
                    continue;
                }

                if (string.IsNullOrEmpty(curr.Value))
                {
                    continue;
                }

                if (string.Compare(curr.Key.ToLower(), "access-control-request-headers") == 0)
                {
                    requestedHeaders = curr.Value.Split(',');
                    break;
                }
            }
        }

        var headers = "";

        if (requestedHeaders != null)
        {
            var addedCount = 0;
            foreach (var curr in requestedHeaders)
            {
                if (string.IsNullOrEmpty(curr))
                {
                    continue;
                }

                if (addedCount > 0)
                {
                    headers += ", ";
                }

                headers += ", " + curr;
                addedCount++;
            }
        }

        foreach (var header in _Settings.Headers)
        {
            context.Response.Headers.Add(header.Key, header.Value);
        }

        context.Response.ContentLength = 0;
        await context.Response.Send().ConfigureAwait(false);
    }
}